import abc
import functools
import inspect
import logging
import time
import types

VALID_TRACE_FLAGS = {'method', 'api'}
TRACE_API = False
TRACE_METHOD = False
TRACE_ENABLED = False
LOG = logging.getLogger('APRSD')


def trace(*dec_args, **dec_kwargs):
    """Trace calls to the decorated function.

    This decorator should always be defined as the outermost decorator so it
    is defined last. This is important so it does not interfere
    with other decorators.

    Using this decorator on a function will cause its execution to be logged at
    `DEBUG` level with arguments, return values, and exceptions.

    :returns: a function decorator
    """

    def _decorator(f):
        func_name = f.__qualname__
        func_file = '/'.join(f.__code__.co_filename.split('/')[-4:])
        func_line_number = f.__code__.co_firstlineno

        @functools.wraps(f)
        def trace_logging_wrapper(*args, **kwargs):
            filter_function = dec_kwargs.get('filter_function')
            logger = LOG

            # NOTE(ameade): Don't bother going any further if DEBUG log level
            # is not enabled for the logger.
            if not logger.isEnabledFor(logging.DEBUG) or not TRACE_ENABLED:
                return f(*args, **kwargs)

            all_args = inspect.getcallargs(f, *args, **kwargs)
            pass_filter = filter_function is None or filter_function(all_args)

            if pass_filter:
                logger.debug(
                    '==> %(func)s: call %(all_args)r file: %(file)s:%(line_number)d',
                    {
                        'func': func_name,
                        'all_args': str(all_args),
                        'file': func_file,
                        'line_number': func_line_number,
                    },
                )

            start_time = time.time() * 1000
            try:
                result = f(*args, **kwargs)
            except Exception as exc:
                total_time = int(round(time.time() * 1000)) - start_time
                logger.debug(
                    '<== %(func)s: exception (%(time)dms) %(exc)r',
                    {
                        'func': func_name,
                        'time': total_time,
                        'exc': exc,
                    },
                )
                raise
            total_time = int(round(time.time() * 1000)) - start_time

            if isinstance(result, dict):
                mask_result = result
            elif isinstance(result, str):
                mask_result = result
            else:
                mask_result = result

            if pass_filter:
                logger.debug(
                    '<== %(func)s: return (%(time)dms) %(result)r',
                    {
                        'func': func_name,
                        'time': total_time,
                        'result': mask_result,
                    },
                )
            return result

        return trace_logging_wrapper

    if len(dec_args) == 0:
        # filter_function is passed and args does not contain f
        return _decorator
    else:
        # filter_function is not passed
        return _decorator(dec_args[0])


def trace_api(*dec_args, **dec_kwargs):
    """Decorates a function if TRACE_API is true."""

    def _decorator(f):
        @functools.wraps(f)
        def trace_api_logging_wrapper(*args, **kwargs):
            if TRACE_API:
                return trace(f, *dec_args, **dec_kwargs)(*args, **kwargs)
            return f(*args, **kwargs)

        return trace_api_logging_wrapper

    if len(dec_args) == 0:
        # filter_function is passed and args does not contain f
        return _decorator
    else:
        # filter_function is not passed
        return _decorator(dec_args[0])


def trace_method(f):
    """Decorates a function if TRACE_METHOD is true."""

    @functools.wraps(f)
    def trace_method_logging_wrapper(*args, **kwargs):
        if TRACE_METHOD:
            return trace(f)(*args, **kwargs)
        return f(*args, **kwargs)

    return trace_method_logging_wrapper


class TraceWrapperMetaclass(type):
    """Metaclass that wraps all methods of a class with trace_method.

    This metaclass will cause every function inside of the class to be
    decorated with the trace_method decorator.

    To use the metaclass you define a class like so:
    class MyClass(object, metaclass=utils.TraceWrapperMetaclass):
    """

    def __new__(cls, classname, bases, class_dict):
        new_class_dict = {}
        for attribute_name, attribute in class_dict.items():
            if isinstance(attribute, types.FunctionType):
                # replace it with a wrapped version
                attribute = functools.update_wrapper(
                    trace_method(attribute),
                    attribute,
                )
            new_class_dict[attribute_name] = attribute

        return type.__new__(cls, classname, bases, new_class_dict)


class TraceWrapperWithABCMetaclass(abc.ABCMeta, TraceWrapperMetaclass):
    """Metaclass that wraps all methods of a class with trace."""


def setup_tracing(trace_flags):
    """Set global variables for each trace flag.

    Sets variables TRACE_METHOD and TRACE_API, which represent
    whether to log methods or api traces.

    :param trace_flags: a list of strings
    """
    global TRACE_METHOD
    global TRACE_API
    global TRACE_ENABLED

    try:
        trace_flags = [flag.strip() for flag in trace_flags]
    except TypeError:  # Handle when trace_flags is None or a test mock
        trace_flags = []
    for invalid_flag in set(trace_flags) - VALID_TRACE_FLAGS:
        LOG.warning('Invalid trace flag: %s', invalid_flag)
    TRACE_METHOD = 'method' in trace_flags
    TRACE_API = 'api' in trace_flags
    TRACE_ENABLED = True



#
#  Dev.py is used to help develop plugins
#
#
# python included libs
import logging
import sys

import click
from oslo_config import cfg

import aprsd
from aprsd import cli_helper, conf, packets, plugin, utils

# local imports here
from aprsd.main import cli
from aprsd.utils import trace

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@cli.group(help='Development type subcommands', context_settings=CONTEXT_SETTINGS)
@click.pass_context
def dev(ctx):
    pass


@dev.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    '--aprs-login',
    envvar='APRS_LOGIN',
    show_envvar=True,
    help='What callsign to send the message from.',
)
@click.option(
    '-p',
    '--plugin',
    'plugin_path',
    show_default=True,
    default=None,
    help='The plugin to run.  Ex: aprsd.plugins.ping.PingPlugin',
)
@click.option(
    '-a',
    '--all',
    'load_all',
    show_default=True,
    is_flag=True,
    default=False,
    help='Load all the plugins in config?',
)
@click.option(
    '-n',
    '--num',
    'number',
    show_default=True,
    default=1,
    help='Number of times to call the plugin',
)
@click.argument('message', nargs=-1, required=True)
@click.pass_context
@cli_helper.process_standard_options
def test_plugin(
    ctx,
    aprs_login,
    plugin_path,
    load_all,
    number,
    message,
):
    """Test an individual APRSD plugin given a python path."""

    LOG.info(f'Python version: {sys.version}')
    LOG.info(f'APRSD DEV Started version: {aprsd.__version__}')
    utils.package.log_installed_extensions_and_plugins()
    CONF.log_opt_values(LOG, logging.DEBUG)

    if not aprs_login:
        if CONF.aprs_network.login == conf.client.DEFAULT_LOGIN:
            click.echo('Must set --aprs_login or APRS_LOGIN')
            ctx.exit(-1)
            return
        else:
            fromcall = CONF.aprs_network.login
    else:
        fromcall = aprs_login

    if not plugin_path:
        click.echo(ctx.get_help())
        click.echo('')
        click.echo('Failed to provide -p option to test a plugin')
        ctx.exit(-1)
        return

    if type(message) is tuple:
        message = ' '.join(message)

    if CONF.trace_enabled:
        trace.setup_tracing(['method', 'api'])

    pm = plugin.PluginManager()
    if load_all:
        pm.setup_plugins(load_help_plugin=CONF.load_help_plugin)
    obj = pm._create_class(plugin_path, plugin.APRSDPluginBase)
    if not obj:
        click.echo(ctx.get_help())
        click.echo('')
        ctx.fail(f"Failed to create object from plugin path '{plugin_path}'")
        ctx.exit()

    # Register the plugin they wanted tested.
    LOG.info(
        'Testing plugin {} Version {}'.format(
            obj.__class__,
            obj.version,
        ),
    )
    pm.register_msg(obj)

    packet = packets.MessagePacket(
        from_call=fromcall,
        to_call=CONF.callsign,
        msgNo=1,
        message_text=message,
    )
    LOG.info(f"P'{plugin_path}'  F'{fromcall}'   C'{message}'")

    for _ in range(number):
        replies = pm.run(packet)
        # Plugin might have threads, so lets stop them so we can exit.
        # obj.stop_threads()
        for reply in replies:
            if isinstance(reply, list):
                # one of the plugins wants to send multiple messages
                for subreply in reply:
                    if isinstance(subreply, packets.Packet):
                        LOG.info(subreply)
                    else:
                        LOG.info(
                            packets.MessagePacket(
                                from_call=CONF.callsign,
                                to_call=fromcall,
                                message_text=subreply,
                            ),
                        )
            elif isinstance(reply, packets.Packet):
                # We have a message based object.
                LOG.info(reply)
            elif reply is not packets.NULL_MESSAGE:
                LOG.info(
                    packets.MessagePacket(
                        from_call=CONF.callsign,
                        to_call=fromcall,
                        message_text=reply,
                    ),
                )
    pm.stop()



import logging
from collections import OrderedDict

from oslo_config import cfg

from aprsd.packets import core
from aprsd.utils import objectstore

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


class PacketList(objectstore.ObjectStoreMixin):
    """Class to keep track of the packets we tx/rx."""

    _instance = None
    _total_rx: int = 0
    _total_tx: int = 0
    maxlen: int = 100

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.maxlen = CONF.packet_list_maxlen
            cls._instance._init_data()
        return cls._instance

    def _init_data(self):
        self.data = {
            'types': {},
            'packets': OrderedDict(),
        }

    def rx(self, packet: type[core.Packet]):
        """Add a packet that was received."""
        with self.lock:
            self._total_rx += 1
            self._add(packet)
            ptype = packet.__class__.__name__
            type_stats = self.data['types'].setdefault(
                ptype,
                {'tx': 0, 'rx': 0},
            )
            type_stats['rx'] += 1

    def tx(self, packet: type[core.Packet]):
        """Add a packet that was received."""
        with self.lock:
            self._total_tx += 1
            self._add(packet)
            ptype = packet.__class__.__name__
            type_stats = self.data['types'].setdefault(
                ptype,
                {'tx': 0, 'rx': 0},
            )
            type_stats['tx'] += 1

    def add(self, packet):
        with self.lock:
            self._add(packet)

    def _add(self, packet):
        if not self.data.get('packets'):
            self._init_data()
        if packet.key in self.data['packets']:
            self.data['packets'].move_to_end(packet.key)
        elif len(self.data['packets']) == self.maxlen:
            self.data['packets'].popitem(last=False)
        self.data['packets'][packet.key] = packet

    def find(self, packet):
        with self.lock:
            return self.data['packets'][packet.key]

    def __len__(self):
        with self.lock:
            return len(self.data['packets'])

    def total_rx(self):
        with self.lock:
            return self._total_rx

    def total_tx(self):
        with self.lock:
            return self._total_tx

    def stats(self, serializable=False) -> dict:
        with self.lock:
            # Get last N packets directly using list slicing
            if CONF.packet_list_stats_maxlen >= 0:
                packets_list = list(self.data.get('packets', {}).values())
                pkts = packets_list[-CONF.packet_list_stats_maxlen :][::-1]
            else:
                # We have to copy here, because this get() results in a pointer
                # to the packets internally here, which can change after this
                # function returns, which would cause a problem trying to save
                # the stats to disk.
                pkts = self.data.get('packets', {}).copy()
            stats = {
                'total_tracked': self._total_rx
                + self._total_tx,  # Fixed typo: was rx + rx
                'rx': self._total_rx,
                'tx': self._total_tx,
                'types': self.data.get('types', {}),  # Changed default from [] to {}
                'packet_count': len(self.data.get('packets', [])),
                'maxlen': self.maxlen,
                'packets': pkts,
            }
            return stats



import queue

# Make these available to anyone importing
# aprsd.threads
from .aprsd import APRSDThread, APRSDThreadList  # noqa: F401
from .rx import (  # noqa: F401
    APRSDProcessPacketThread,
    APRSDRXThread,
)

packet_queue = queue.Queue(maxsize=500)



#  Utilities for plugins to use
import json
import logging

import requests

LOG = logging.getLogger('APRSD')


def get_aprs_fi(api_key, callsign):
    LOG.debug(f"Fetch aprs.fi location for '{callsign}'")
    try:
        url = (
            'http://api.aprs.fi/api/get?&what=loc&apikey={}&format=json&name={}'.format(
                api_key, callsign
            )
        )
        response = requests.get(url)
    except Exception as e:
        raise Exception('Failed to get aprs.fi location') from e
    else:
        response.raise_for_status()
        return json.loads(response.text)


def get_weather_gov_for_gps(lat, lon):
    # FIXME(hemna) This is currently BROKEN
    LOG.debug(f'Fetch station at {lat}, {lon}')
    headers = requests.utils.default_headers()
    headers.update(
        {'User-Agent': '(aprsd, waboring@hemna.com)'},
    )
    try:
        url2 = (
            'https://forecast.weather.gov/MapClick.php?lat=%s'
            '&lon=%s&FcstType=json' % (lat, lon)
            # f"https://api.weather.gov/points/{lat},{lon}"
        )
        LOG.debug(f"Fetching weather '{url2}'")
        response = requests.get(url2, headers=headers)
    except Exception as e:
        LOG.error(e)
        raise Exception('Failed to get weather') from e
    else:
        response.raise_for_status()
        return json.loads(response.text)


def get_weather_gov_metar(station):
    LOG.debug(f"Fetch metar for station '{station}'")
    try:
        url = 'https://api.weather.gov/stations/{}/observations/latest'.format(
            station,
        )
        response = requests.get(url)
    except Exception as e:
        raise Exception('Failed to fetch metar') from e
    else:
        response.raise_for_status()
        return json.loads(response)


def fetch_openweathermap(api_key, lat, lon, units='metric', exclude=None):
    LOG.debug(f'Fetch openweathermap for {lat}, {lon}')
    if not exclude:
        exclude = 'minutely,hourly,daily,alerts'
    try:
        url = (
            "https://api.openweathermap.org/data/3.0/onecall?"
            "lat={}&lon={}&appid={}&units={}&exclude={}".format(
                lat,
                lon,
                api_key,
                units,
                exclude,
            )
        )
        LOG.debug(f"Fetching OWM weather '{url}'")
        response = requests.get(url)
    except Exception as e:
        LOG.error(e)
        raise Exception('Failed to get weather') from e
    else:
        response.raise_for_status()
        return json.loads(response.text)



import unittest
from unittest import mock

import aprslib
from aprslib import util as aprslib_util

from aprsd import packets
from aprsd.packets import core

from . import fake


class TestPacketBase(unittest.TestCase):
    def _fake_dict(
        self,
        from_call=fake.FAKE_FROM_CALLSIGN,
        to_call=fake.FAKE_TO_CALLSIGN,
        message=None,
        msg_number=None,
        message_format=core.PACKET_TYPE_MESSAGE,
    ):
        packet_dict = {
            "from": from_call,
            "addresse": to_call,
            "to": to_call,
            "format": message_format,
            "raw": "",
        }

        if message:
            packet_dict["message_text"] = message

        if msg_number:
            packet_dict["msgNo"] = str(msg_number)

        return packet_dict

    def test_packet_construct(self):
        pkt = packets.Packet(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
        )

        self.assertEqual(fake.FAKE_FROM_CALLSIGN, pkt.from_call)
        self.assertEqual(fake.FAKE_TO_CALLSIGN, pkt.to_call)

    def test_packet_get_attr(self):
        pkt = packets.Packet(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
        )

        self.assertEqual(
            fake.FAKE_FROM_CALLSIGN,
            pkt.get("from_call"),
        )

    def test_packet_factory(self):
        pkt_dict = self._fake_dict()
        pkt = packets.factory(pkt_dict)

        self.assertIsInstance(pkt, packets.MessagePacket)
        self.assertEqual(fake.FAKE_FROM_CALLSIGN, pkt.from_call)
        self.assertEqual(fake.FAKE_TO_CALLSIGN, pkt.to_call)
        self.assertEqual(fake.FAKE_TO_CALLSIGN, pkt.addresse)

        pkt_dict["symbol"] = "_"
        pkt_dict["weather"] = {
            "wind_gust": 1.11,
            "temperature": 32.01,
            "humidity": 85,
            "pressure": 1095.12,
            "comment": "Home!",
        }
        pkt_dict["format"] = core.PACKET_TYPE_UNCOMPRESSED
        pkt = packets.factory(pkt_dict)
        self.assertIsInstance(pkt, packets.WeatherPacket)

    @mock.patch("aprsd.packets.core.GPSPacket._build_time_zulu")
    def test_packet_format_rain_1h(self, mock_time_zulu):
        mock_time_zulu.return_value = "221450"

        wx = packets.WeatherPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            timestamp=1671721164.1112509,
        )
        wx.prepare()

        expected = "KFAKE>KMINE,WIDE1-1,WIDE2-1:@221450z0.0/0.0_000/000g000t000r000p000P000h00b00000"
        self.assertEqual(expected, wx.raw)
        rain_location = 59
        self.assertEqual(rain_location, wx.raw.find("r000"))

        wx.rain_1h = 1.11
        wx.prepare()
        expected = "KFAKE>KMINE,WIDE1-1,WIDE2-1:@221450z0.0/0.0_000/000g000t000r111p000P000h00b00000"
        self.assertEqual(expected, wx.raw)

        wx.rain_1h = 0.01
        wx.prepare()
        expected = "KFAKE>KMINE,WIDE1-1,WIDE2-1:@221450z0.0/0.0_000/000g000t000r001p000P000h00b00000"
        self.assertEqual(expected, wx.raw)

    def test_beacon_factory(self):
        """Test to ensure a beacon packet is created."""
        packet_raw = (
            "WB4BOR-12>APZ100,WIDE2-1:@161647z3724.15N107847.58W$ APRSD WebChat"
        )
        packet_dict = aprslib.parse(packet_raw)
        packet = packets.factory(packet_dict)
        self.assertIsInstance(packet, packets.BeaconPacket)

        packet_raw = "kd8mey-10>APRS,TCPIP*,qAC,T2SYDNEY:=4247.80N/08539.00WrPHG1210/Making 220 Great Again Allstar# 552191"
        packet_dict = aprslib.parse(packet_raw)
        packet = packets.factory(packet_dict)
        self.assertIsInstance(packet, packets.BeaconPacket)

    def test_reject_factory(self):
        """Test to ensure a reject packet is created."""
        packet_raw = "HB9FDL-1>APK102,HB9FM-4*,WIDE2,qAR,HB9FEF-11::REPEAT   :rej4139"
        packet_dict = aprslib.parse(packet_raw)
        packet = packets.factory(packet_dict)
        self.assertIsInstance(packet, packets.RejectPacket)

        self.assertEqual("4139", packet.msgNo)
        self.assertEqual("HB9FDL-1", packet.from_call)
        self.assertEqual("REPEAT", packet.to_call)
        self.assertEqual("reject", packet.packet_type)
        self.assertIsNone(packet.payload)

    def test_thirdparty_factory(self):
        """Test to ensure a third party packet is created."""
        packet_raw = "GTOWN>APDW16,WIDE1-1,WIDE2-1:}KM6LYW-9>APZ100,TCPIP,GTOWN*::KM6LYW   :KM6LYW: 19 Miles SW"
        packet_dict = aprslib.parse(packet_raw)
        packet = packets.factory(packet_dict)
        self.assertIsInstance(packet, packets.ThirdPartyPacket)

    def test_weather_factory(self):
        """Test to ensure a weather packet is created."""
        packet_raw = "FW9222>APRS,TCPXX*,qAX,CWOP-6:@122025z2953.94N/08423.77W_232/003g006t084r000p032P000h80b10157L745.DsWLL"
        packet_dict = aprslib.parse(packet_raw)
        packet = packets.factory(packet_dict)
        self.assertIsInstance(packet, packets.WeatherPacket)

        self.assertEqual(28.88888888888889, packet.temperature)
        self.assertEqual(0.0, packet.rain_1h)
        self.assertEqual(1015.7, packet.pressure)
        self.assertEqual(80, packet.humidity)
        self.assertEqual(745, packet.luminosity)
        self.assertEqual(3.0, packet.wind_speed)
        self.assertEqual(232, packet.wind_direction)
        self.assertEqual(6.0, packet.wind_gust)
        self.assertEqual(29.899, packet.latitude)
        self.assertEqual(-84.39616666666667, packet.longitude)

    def test_mice_factory(self):
        packet_raw = 'kh2sr-15>S7TSYR,WIDE1-1,WIDE2-1,qAO,KO6KL-1:`1`7\x1c\x1c.#/`"4,}QuirkyQRP 4.6V  35.3C S06'
        packet_dict = aprslib.parse(packet_raw)
        packet = packets.factory(packet_dict)
        self.assertIsInstance(packet, packets.MicEPacket)

        # Packet with telemetry and DAO
        # http://www.aprs.org/datum.txt
        packet_raw = (
            "KD9YIL>T0PX9W,WIDE1-1,WIDE2-1,qAO,NU9R-10:`sB,l#P>/'\"6+}|#*%U'a|!whl!|3"
        )
        packet_dict = aprslib.parse(packet_raw)
        packet = packets.factory(packet_dict)
        self.assertIsInstance(packet, packets.MicEPacket)

    def test_ack_format(self):
        """Test the ack packet format."""
        ack = packets.AckPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            msgNo=123,
        )

        expected = (
            f"{fake.FAKE_FROM_CALLSIGN}>APZ100::{fake.FAKE_TO_CALLSIGN:<9}:ack123"
        )
        self.assertEqual(expected, str(ack))

    def test_reject_format(self):
        """Test the reject packet format."""
        reject = packets.RejectPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            msgNo=123,
        )

        expected = (
            f"{fake.FAKE_FROM_CALLSIGN}>APZ100::{fake.FAKE_TO_CALLSIGN:<9}:rej123"
        )
        self.assertEqual(expected, str(reject))

    def test_beacon_format(self):
        """Test the beacon packet format."""
        lat = 28.123456
        lon = -80.123456
        ts = 1711219496.6426
        comment = "My Beacon Comment"
        packet = packets.BeaconPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            latitude=lat,
            longitude=lon,
            timestamp=ts,
            symbol=">",
            comment=comment,
        )

        expected_lat = aprslib_util.latitude_to_ddm(lat)
        expected_lon = aprslib_util.longitude_to_ddm(lon)
        expected = f"KFAKE>APZ100:@231844z{expected_lat}/{expected_lon}>{comment}"
        self.assertEqual(expected, str(packet))

    def test_beacon_format_no_comment(self):
        """Test the beacon packet format."""
        lat = 28.123456
        lon = -80.123456
        ts = 1711219496.6426
        packet = packets.BeaconPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            latitude=lat,
            longitude=lon,
            timestamp=ts,
            symbol=">",
        )
        empty_comment = "APRSD Beacon"

        expected_lat = aprslib_util.latitude_to_ddm(lat)
        expected_lon = aprslib_util.longitude_to_ddm(lon)
        expected = f"KFAKE>APZ100:@231844z{expected_lat}/{expected_lon}>{empty_comment}"
        self.assertEqual(expected, str(packet))

    def test_bulletin_format(self):
        """Test the bulletin packet format."""
        # bulletin id = 0
        bid = 0
        packet = packets.BulletinPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            message_text="My Bulletin Message",
            bid=0,
        )

        expected = (
            f"{fake.FAKE_FROM_CALLSIGN}>APZ100::BLN{bid:<9}:{packet.message_text}"
        )
        self.assertEqual(expected, str(packet))

        # bulletin id = 1
        bid = 1
        txt = "((((((( CX2SA - Salto Uruguay ))))))) http://www.cx2sa.org"
        packet = packets.BulletinPacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            message_text=txt,
            bid=1,
        )

        expected = f"{fake.FAKE_FROM_CALLSIGN}>APZ100::BLN{bid:<9}:{txt}"
        self.assertEqual(expected, str(packet))

    def test_message_format(self):
        """Test the message packet format."""

        message = "My Message"
        msgno = "ABX"
        packet = packets.MessagePacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            message_text=message,
            msgNo=msgno,
        )

        expected = f"{fake.FAKE_FROM_CALLSIGN}>APZ100::{fake.FAKE_TO_CALLSIGN:<9}:{message}{{{msgno}"
        self.assertEqual(expected, str(packet))

        # test with bad words
        # Currently fails with mixed case
        message = "My cunt piss fuck shIt text"
        exp_msg = "My **** **** **** **** text"
        msgno = "ABX"
        packet = packets.MessagePacket(
            from_call=fake.FAKE_FROM_CALLSIGN,
            to_call=fake.FAKE_TO_CALLSIGN,
            message_text=message,
            msgNo=msgno,
        )
        expected = f"{fake.FAKE_FROM_CALLSIGN}>APZ100::{fake.FAKE_TO_CALLSIGN:<9}:{exp_msg}{{{msgno}"
        self.assertEqual(expected, str(packet))



import logging
from typing import Callable, Protocol, runtime_checkable

from aprsd.utils import singleton

LOG = logging.getLogger('APRSD')


@runtime_checkable
class StatsProducer(Protocol):
    """The StatsProducer protocol is used to define the interface for collecting stats."""

    def stats(self, serializable=False) -> dict:
        """provide stats in a dictionary format."""
        ...


@singleton
class Collector:
    """The Collector class is used to collect stats from multiple StatsProducer instances."""

    def __init__(self):
        self.producers: list[Callable] = []

    def collect(self, serializable=False) -> dict:
        stats = {}
        for name in self.producers:
            cls = name()
            try:
                stats[cls.__class__.__name__] = cls.stats(
                    serializable=serializable
                ).copy()
            except Exception as e:
                LOG.error(f'Error in producer {name} (stats): {e}')
                raise e
        return stats

    def register_producer(self, producer_name: Callable):
        if not isinstance(producer_name, StatsProducer):
            raise TypeError(f'Producer {producer_name} is not a StatsProducer')
        self.producers.append(producer_name)

    def unregister_producer(self, producer_name: Callable):
        if not isinstance(producer_name, StatsProducer):
            raise TypeError(f'Producer {producer_name} is not a StatsProducer')
        self.producers.remove(producer_name)

    def stop_all(self):
        """Stop and unregister all registered stats producers."""
        for producer in self.producers[:]:
            LOG.info(f'Stopping Stats producer {producer}')
            self.unregister_producer(producer)



import logging
import time

from aprsd import plugin
from aprsd.utils import trace


LOG = logging.getLogger("APRSD")


class PingPlugin(plugin.APRSDRegexCommandPluginBase):
    """Ping."""

    command_regex = r"^([p]|[p]\s|ping)"
    command_name = "ping"
    short_description = "reply with a Pong!"

    @trace.trace
    def process(self, packet):
        LOG.info("PingPlugin")
        # fromcall = packet.get("from")
        # message = packet.get("message_text", None)
        # ack = packet.get("msgNo", "0")
        stm = time.localtime()
        h = stm.tm_hour
        m = stm.tm_min
        s = stm.tm_sec
        reply = (
            "Pong! " + str(h).zfill(2) + ":" + str(m).zfill(2) + ":" + str(s).zfill(2)
        )
        return reply.rstrip()



import logging
import os
import pathlib
import pickle
import threading

from oslo_config import cfg

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')


class ObjectStoreMixin:
    """Class 'MIXIN' intended to save/load object data.

    The asumption of how this mixin is used:
      The using class has to have a:
         * data in self.data as a dictionary
         * a self.lock thread lock
         * Class must specify self.save_file as the location.


    When APRSD quits, it calls save()
    When APRSD Starts, it calls load()
    aprsd server -f (flush) will wipe all saved objects.
    """

    def __init__(self):
        self.lock = threading.RLock()

    def __len__(self):
        with self.lock:
            return len(self.data)

    def __iter__(self):
        with self.lock:
            return iter(self.data)

    def get_all(self):
        with self.lock:
            return self.data

    def get(self, key):
        with self.lock:
            return self.data.get(key)

    def copy(self):
        with self.lock:
            return self.data.copy()

    def _init_store(self):
        if not CONF.enable_save:
            return
        sl = CONF.save_location
        if not os.path.exists(sl):
            LOG.warning(f"Save location {sl} doesn't exist")
            try:
                os.makedirs(sl)
            except Exception as ex:
                LOG.exception(ex)

    def _save_filename(self):
        save_location = CONF.save_location

        return '{}/{}.p'.format(
            save_location,
            self.__class__.__name__.lower(),
        )

    def save(self):
        """Save any queued to disk?"""
        if not CONF.enable_save:
            return
        self._init_store()
        save_filename = self._save_filename()
        if len(self) > 0:
            LOG.debug(
                f'{self.__class__.__name__}::Saving'
                f' {len(self)} entries to disk at '
                f'{save_filename}',
            )
            with self.lock:
                with open(save_filename, 'wb+') as fp:
                    pickle.dump(self.data, fp)
        else:
            LOG.debug(
                "{} Nothing to save, flushing old save file '{}'".format(
                    self.__class__.__name__,
                    save_filename,
                ),
            )
            self.flush()

    def load(self):
        if not CONF.enable_save:
            return
        if os.path.exists(self._save_filename()):
            try:
                with open(self._save_filename(), 'rb') as fp:
                    raw = pickle.load(fp)
                    if raw:
                        self.data = raw
                        LOG.debug(
                            f'{self.__class__.__name__}::Loaded {len(self)} entries from disk.',
                        )
                    else:
                        LOG.debug(f'{self.__class__.__name__}::No data to load.')
            except (pickle.UnpicklingError, Exception) as ex:
                LOG.error(f'Failed to UnPickle {self._save_filename()}')
                LOG.error(ex)
                self.data = {}
        else:
            LOG.debug(f'{self.__class__.__name__}::No save file found.')

    def flush(self):
        """Nuke the old pickle file that stored the old results from last aprsd run."""
        if not CONF.enable_save:
            return
        if os.path.exists(self._save_filename()):
            pathlib.Path(self._save_filename()).unlink()
        with self.lock:
            self.data = {}



from __future__ import annotations

import abc
import importlib
import inspect
import logging
import re
import textwrap
import threading

import pluggy
from oslo_config import cfg

import aprsd
from aprsd import packets, threads
from aprsd.client.client import APRSDClient
from aprsd.packets import watch_list

# setup the global logger
CONF = cfg.CONF
LOG = logging.getLogger('APRSD')

CORE_MESSAGE_PLUGINS = [
    'aprsd.plugins.email.EmailPlugin',
    'aprsd.plugins.fortune.FortunePlugin',
    'aprsd.plugins.location.LocationPlugin',
    'aprsd.plugins.ping.PingPlugin',
    'aprsd.plugins.time.TimePlugin',
    'aprsd.plugins.weather.USWeatherPlugin',
    'aprsd.plugins.version.VersionPlugin',
]

CORE_NOTIFY_PLUGINS = [
    'aprsd.plugins.notify.NotifySeenPlugin',
]

hookspec = pluggy.HookspecMarker('aprsd')
hookimpl = pluggy.HookimplMarker('aprsd')


class APRSDPluginSpec:
    """A hook specification namespace."""

    @hookspec
    def filter(self, packet: type[packets.Packet]):
        """My special little hook that you can customize."""


class APRSDPluginBase(metaclass=abc.ABCMeta):
    """The base class for all APRSD Plugins."""

    config = None
    rx_count = 0
    tx_count = 0
    version = aprsd.__version__

    # Holds the list of APRSDThreads that the plugin creates
    threads = []
    # Set this in setup()
    enabled = False

    def __init__(self):
        self.message_counter = 0
        self.setup()
        self.threads = self.create_threads() or []
        self.start_threads()

    def start_threads(self) -> None:
        if self.enabled and self.threads:
            if not isinstance(self.threads, list):
                self.threads = [self.threads]

            try:
                for thread in self.threads:
                    if isinstance(thread, threads.APRSDThread):
                        thread.start()
                    else:
                        LOG.error(
                            "Can't start thread {}:{}, Must be a child "
                            'of aprsd.threads.APRSDThread'.format(
                                self,
                                thread,
                            ),
                        )
            except Exception:
                LOG.error(
                    'Failed to start threads for plugin {}'.format(
                        self,
                    ),
                )

    @property
    def message_count(self) -> int:
        return self.message_counter

    def help(self) -> str:
        return 'Help!'

    @abc.abstractmethod
    def setup(self):
        """Do any plugin setup here."""
        self.enabled = True

    def create_threads(self):
        """Gives the plugin writer the ability start a background thread."""
        return []

    def rx_inc(self):
        self.rx_count += 1

    def tx_inc(self):
        self.tx_count += 1

    def stop_threads(self):
        """Stop any threads this plugin might have created."""
        for thread in self.threads:
            if isinstance(thread, threads.APRSDThread):
                thread.stop()

    @abc.abstractmethod
    def filter(self, packet: type[packets.Packet]) -> str | packets.MessagePacket:
        pass

    @abc.abstractmethod
    def process(self, packet: type[packets.Packet]):
        """This is called when the filter passes."""


class APRSDWatchListPluginBase(APRSDPluginBase, metaclass=abc.ABCMeta):
    """Base plugin class for all notification APRSD plugins.

    All these plugins will get every packet seen by APRSD's
    registered list of HAM callsigns in the config file's
    watch_list.

    When you want to 'notify' something when a packet is seen
    by a particular HAM callsign, write a plugin based off of
    this class.
    """

    def setup(self):
        # if we have a watch list enabled, we need to add filtering
        # to enable seeing packets from the watch list.
        if CONF.watch_list.enabled:
            # watch list is enabled
            self.enabled = True
            watch_list = CONF.watch_list.callsigns
            # make sure the timeout is set or this doesn't work
            if watch_list:
                aprs_client = APRSDClient()
                filter_str = 'b/{}'.format('/'.join(watch_list))
                aprs_client.set_filter(filter_str)
            else:
                LOG.warning('Watch list enabled, but no callsigns set.')

    @hookimpl
    def filter(self, packet: type[packets.Packet]) -> str | packets.MessagePacket:
        result = packets.NULL_MESSAGE
        if self.enabled:
            wl = watch_list.WatchList()
            if wl.callsign_in_watchlist(packet.from_call):
                # packet is from a callsign in the watch list
                self.rx_inc()
                try:
                    result = self.process(packet)
                except Exception as ex:
                    LOG.error(
                        'Plugin {} failed to process packet {}'.format(
                            self.__class__,
                            ex,
                        ),
                    )
                if result:
                    self.tx_inc()
        else:
            LOG.warning(f'{self.__class__} plugin is not enabled')

        return result


class APRSDRegexCommandPluginBase(APRSDPluginBase, metaclass=abc.ABCMeta):
    """Base Message plugin class.

    When you want to search for a particular command in an
    APRSD message and send a direct reply, write a plugin
    based off of this class.
    """

    @property
    def command_name(self):
        """The usage string help."""
        raise NotImplementedError

    @property
    def command_regex(self):
        """The regex to match from the caller"""
        raise NotImplementedError

    def help(self):
        return '{}: {}'.format(
            self.command_name.lower(),
            self.command_regex,
        )

    def setup(self):
        """Do any plugin setup here."""
        self.enabled = True

    @hookimpl
    def filter(self, packet: packets.MessagePacket) -> str | packets.MessagePacket:
        LOG.debug(f'{self.__class__.__name__} called')
        if not self.enabled:
            result = f"{self.__class__.__name__} isn't enabled"
            LOG.warning(result)
            return result

        if not isinstance(packet, packets.MessagePacket):
            LOG.warning(
                f'{self.__class__.__name__} Got a {packet.__class__.__name__} ignoring'
            )
            return packets.NULL_MESSAGE

        result = None

        message = packet.message_text
        tocall = packet.to_call

        # Only process messages destined for us
        # and is an APRS message format and has a message.
        if (
            tocall == CONF.callsign
            and isinstance(packet, packets.MessagePacket)
            and message
        ):
            if re.search(self.command_regex, message, re.IGNORECASE):
                self.rx_inc()
                try:
                    result = self.process(packet)
                except Exception as ex:
                    LOG.error(
                        'Plugin {} failed to process packet {}'.format(
                            self.__class__,
                            ex,
                        ),
                    )
                    LOG.exception(ex)
                if result:
                    self.tx_inc()

        return result


class APRSFIKEYMixin:
    """Mixin class to enable checking the existence of the aprs.fi apiKey."""

    def ensure_aprs_fi_key(self):
        if not CONF.aprs_fi.apiKey:
            LOG.error('Config aprs_fi.apiKey is not set')
            self.enabled = False
        else:
            self.enabled = True


class HelpPlugin(APRSDRegexCommandPluginBase):
    """Help Plugin that is always enabled.

    This plugin is in this file to prevent a circular import.
    """

    command_regex = '^[hH]'
    command_name = 'help'

    def help(self):
        return 'Help: send APRS help or help <plugin>'

    def process(self, packet: packets.MessagePacket):
        LOG.info('HelpPlugin')
        # fromcall = packet.get("from")
        message = packet.message_text
        # ack = packet.get("msgNo", "0")
        a = re.search(r'^.*\s+(.*)', message)
        command_name = None
        if a is not None:
            command_name = a.group(1).lower()

        pm = PluginManager()

        if command_name and '?' not in command_name:
            # user wants help for a specific plugin
            reply = None
            for p in pm.get_plugins():
                if (
                    p.enabled
                    and isinstance(p, APRSDRegexCommandPluginBase)
                    and p.command_name.lower() == command_name
                ):
                    reply = p.help()

            if reply:
                return reply

        list = []
        for p in pm.get_plugins():
            LOG.debug(p)
            if p.enabled and isinstance(p, APRSDRegexCommandPluginBase):
                name = p.command_name.lower()
                if name not in list and 'help' not in name:
                    list.append(name)

        list.sort()
        reply = ' '.join(list)
        lines = textwrap.wrap(reply, 60)
        replies = ["Send APRS MSG of 'help' or 'help <plugin>'"]
        for line in lines:
            replies.append(f'plugins: {line}')

        for entry in replies:
            LOG.debug(f'{len(entry)} {entry}')

        LOG.debug(f'{replies}')
        return replies


class PluginManager:
    # The singleton instance object for this class
    _instance = None

    # the pluggy PluginManager for all Message plugins
    _pluggy_pm = None
    # the pluggy PluginManager for all WatchList plugins
    _watchlist_pm = None

    lock = None

    def __new__(cls, *args, **kwargs):
        """This magic turns this into a singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # Put any initialization here.
            cls._instance.lock = threading.Lock()
            cls._instance._init()
        return cls._instance

    def _init(self):
        self._pluggy_pm = pluggy.PluginManager('aprsd')
        self._pluggy_pm.add_hookspecs(APRSDPluginSpec)
        # For the watchlist plugins
        self._watchlist_pm = pluggy.PluginManager('aprsd')
        self._watchlist_pm.add_hookspecs(APRSDPluginSpec)

    def stats(self, serializable=False) -> dict:
        """Collect and return stats for all plugins."""

        def full_name_with_qualname(obj):
            return '{}.{}'.format(
                obj.__class__.__module__,
                obj.__class__.__qualname__,
            )

        plugin_stats = {}
        plugins = self.get_plugins()
        if plugins:
            for p in plugins:
                plugin_stats[full_name_with_qualname(p)] = {
                    'enabled': p.enabled,
                    'rx': p.rx_count,
                    'tx': p.tx_count,
                    'version': p.version,
                }

        return plugin_stats

    def is_plugin(self, obj):
        for c in inspect.getmro(obj):
            if issubclass(c, APRSDPluginBase):
                return True

        return False

    def _create_class(
        self,
        module_class_string,
        super_cls: type = None,
        **kwargs,
    ):
        """
        Method to create a class from a fqn python string.
        :param module_class_string: full name of the class to create an object
        :param super_cls: expected super class for validity, None if bypass
        :param kwargs: parameters to pass
        :return:
        """
        module_name = None
        class_name = None
        try:
            module_name, class_name = module_class_string.rsplit('.', 1)
            module = importlib.import_module(module_name)
            # Commented out because the email thread starts in a different context
            # and hence gives a different singleton for the EmailStats
            # module = importlib.reload(module)
        except Exception as ex:
            if not module_name:
                LOG.error(f'Failed to load Plugin {module_class_string}')
            else:
                LOG.error(f"Failed to load Plugin '{module_name}' : '{ex}'")
            return

        assert hasattr(module, class_name), 'class {} is not in {}'.format(
            class_name,
            module_name,
        )
        # click.echo('reading class {} from module {}'.format(
        #     class_name, module_name))
        cls = getattr(module, class_name)
        if super_cls is not None:
            assert issubclass(cls, super_cls), 'class {} should inherit from {}'.format(
                class_name,
                super_cls.__name__,
            )
        # click.echo('initialising {} with params {}'.format(class_name, kwargs))
        obj = cls(**kwargs)
        return obj

    def _load_plugin(self, plugin_name):
        """
        Given a python fully qualified class path.name,
        Try importing the path, then creating the object,
        then registering it as a aprsd Command Plugin
        """
        plugin_obj = None
        try:
            plugin_obj = self._create_class(
                plugin_name,
                APRSDPluginBase,
            )
            if plugin_obj:
                if isinstance(plugin_obj, APRSDWatchListPluginBase):
                    if plugin_obj.enabled:
                        LOG.info(
                            "Registering WatchList plugin '{}'({})".format(
                                plugin_name,
                                plugin_obj.version,
                            ),
                        )
                        self._watchlist_pm.register(plugin_obj)
                    else:
                        LOG.warning(
                            f'Plugin {plugin_obj.__class__.__name__} is disabled'
                        )
                elif isinstance(plugin_obj, APRSDRegexCommandPluginBase):
                    if plugin_obj.enabled:
                        LOG.info(
                            "Registering Regex plugin '{}'({}) -- {}".format(
                                plugin_name,
                                plugin_obj.version,
                                plugin_obj.command_regex,
                            ),
                        )
                        self._pluggy_pm.register(plugin_obj)
                    else:
                        LOG.warning(
                            f'Plugin {plugin_obj.__class__.__name__} is disabled'
                        )
                elif isinstance(plugin_obj, APRSDPluginBase):
                    if plugin_obj.enabled:
                        LOG.info(
                            "Registering Base plugin '{}'({})".format(
                                plugin_name,
                                plugin_obj.version,
                            ),
                        )
                        self._pluggy_pm.register(plugin_obj)
                    else:
                        LOG.warning(
                            f'Plugin {plugin_obj.__class__.__name__} is disabled'
                        )
        except Exception as ex:
            LOG.error(f"Couldn't load plugin '{plugin_name}'")
            LOG.exception(ex)

    def reload_plugins(self):
        with self.lock:
            del self._pluggy_pm
            self.setup_plugins(load_help_plugin=CONF.load_help_plugin)

    def setup_plugins(
        self,
        load_help_plugin=True,
        plugin_list=None,
    ):
        """Create the plugin manager and register plugins."""

        LOG.info('Loading APRSD Plugins')
        # Help plugin is always enabled.
        if load_help_plugin:
            _help = HelpPlugin()
            self._pluggy_pm.register(_help)

        # if plugins_list is passed in, only load
        # those plugins.
        if plugin_list:
            for plugin_name in plugin_list:
                self._load_plugin(plugin_name)
        elif CONF.enabled_plugins:
            for p_name in CONF.enabled_plugins:
                self._load_plugin(p_name)
        else:
            # Enabled plugins isn't set, so we default to loading all of
            # the core plugins.
            for p_name in CORE_MESSAGE_PLUGINS:
                self._load_plugin(p_name)

        LOG.info('Completed Plugin Loading.')

    def run(self, packet: packets.MessagePacket):
        """Execute all the plugins run method."""
        with self.lock:
            return self._pluggy_pm.hook.filter(packet=packet)

    def run_watchlist(self, packet: packets.Packet):
        with self.lock:
            return self._watchlist_pm.hook.filter(packet=packet)

    def stop(self):
        """Stop all threads created by all plugins."""
        with self.lock:
            for p in self.get_plugins():
                if hasattr(p, 'stop_threads'):
                    p.stop_threads()

    def register_msg(self, obj):
        """Register the plugin."""
        with self.lock:
            self._pluggy_pm.register(obj)

    def get_plugins(self):
        plugin_list = []
        if self._pluggy_pm:
            for plug in self._pluggy_pm.get_plugins():
                plugin_list.append(plug)
        if self._watchlist_pm:
            for plug in self._watchlist_pm.get_plugins():
                plugin_list.append(plug)

        return plugin_list

    def get_watchlist_plugins(self):
        pl = []
        if self._watchlist_pm:
            for plug in self._watchlist_pm.get_plugins():
                pl.append(plug)
        return pl

    def get_message_plugins(self):
        pl = []
        if self._pluggy_pm:
            for plug in self._pluggy_pm.get_plugins():
                pl.append(plug)
        return pl



#
#  Used to fetch the stats url and determine if
#  aprsd server is 'healthy'
#
#
# python included libs
import datetime
import logging
import sys

import click
from oslo_config import cfg
from rich.console import Console

import aprsd
from aprsd import (  # noqa: F401
    cli_helper,
    conf,
)

# local imports here
from aprsd.main import cli
from aprsd.threads import stats as stats_threads

# setup the global logger
# log.basicConfig(level=log.DEBUG) # level=10
CONF = cfg.CONF
LOG = logging.getLogger("APRSD")
console = Console()


@cli.command()
@cli_helper.add_options(cli_helper.common_options)
@click.option(
    "--timeout",
    show_default=True,
    default=3,
    help="How long to wait for healtcheck url to come back",
)
@click.pass_context
@cli_helper.process_standard_options
def healthcheck(ctx, timeout):
    """Check the health of the running aprsd server."""
    ver_str = f"APRSD HealthCheck version: {aprsd.__version__}"
    console.log(ver_str)

    with console.status(ver_str):
        try:
            stats_obj = stats_threads.StatsStore()
            stats_obj.load()
            stats = stats_obj.data
            # console.print(stats)
        except Exception as ex:
            console.log(f"Failed to load stats: '{ex}'")
            sys.exit(-1)
        else:
            now = datetime.datetime.now()
            if not stats:
                console.log("No stats from aprsd")
                sys.exit(-1)

            email_stats = stats.get("EmailStats")
            if email_stats:
                email_thread_last_update = email_stats["last_check_time"]

                if email_thread_last_update != "never":
                    d = now - email_thread_last_update
                    max_timeout = {"hours": 0.0, "minutes": 5, "seconds": 30}
                    max_delta = datetime.timedelta(**max_timeout)
                    if d > max_delta:
                        console.log(f"Email thread is very old! {d}")
                        sys.exit(-1)

            client_stats = stats.get("APRSClientStats")
            if not client_stats:
                console.log("No APRSClientStats")
                sys.exit(-1)
            else:
                aprsis_last_update = client_stats["connection_keepalive"]
                d = now - aprsis_last_update
                max_timeout = {"hours": 0.0, "minutes": 5, "seconds": 0}
                max_delta = datetime.timedelta(**max_timeout)
                if d > max_delta:
                    LOG.error(f"APRS-IS last update is very old! {d}")
                    sys.exit(-1)

            console.log("OK")
            sys.exit(0)



import logging
import threading
from typing import Callable

import timeago
import wrapt
from loguru import logger
from oslo_config import cfg

from aprsd.client import drivers  # noqa - ensure drivers are registered
from aprsd.client.drivers.registry import DriverRegistry
from aprsd.packets import core
from aprsd.utils import keepalive_collector, trace

CONF = cfg.CONF
LOG = logging.getLogger('APRSD')
LOGU = logger


class APRSDClient(metaclass=trace.TraceWrapperMetaclass):
    """APRSD client class.

    This is a singleton class that provides a single instance of the APRSD client.
    It is responsible for connecting to the appropriate APRSD client driver based on
    the configuration.

    """

    _instance = None
    driver = None
    lock = threading.Lock()
    filter = None

    def __new__(cls, *args, **kwargs):
        """This magic turns this into a singleton."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            keepalive_collector.KeepAliveCollector().register(cls)
        return cls._instance

    def __init__(self, auto_connect: bool = True):
        self.auto_connect = auto_connect
        self.connected = False
        self.running = False
        self.login_status = {
            'success': False,
            'message': None,
        }
        if not self.driver:
            self.driver = DriverRegistry().get_driver()
        if self.auto_connect:
            self.connect()

    def stats(self, serializable=False) -> dict:
        stats = {}
        if self.driver:
            stats = self.driver.stats(serializable=serializable)
        return stats

    @staticmethod
    def is_enabled():
        for driver in DriverRegistry().drivers:
            if driver.is_enabled():
                return True
        return False

    @staticmethod
    def is_configured():
        """Check if ANY driver is configured."""
        for driver in DriverRegistry().drivers:
            if driver.is_configured():
                return True
        return False

    # @property
    # def is_connected(self):
    #     if not self.driver:
    #         return False
    #     return self.driver.is_connected()

    @property
    def login_success(self):
        if not self.driver:
            return False
        return self.driver.login_success

    @property
    def login_failure(self):
        if not self.driver:
            return None
        return self.driver.login_failure

    def set_filter(self, filter):
        self.filter = filter
        if not self.driver:
            return
        self.driver.set_filter(filter)

    def get_filter(self):
        if not self.driver:
            return None
        return self.driver.filter

    def is_alive(self):
        return self.driver.is_alive

    @wrapt.synchronized(lock)
    def connect(self):
        if not self.driver:
            self.driver = DriverRegistry().get_driver()
        if not self.connected:
            self.driver.setup_connection()
            self.connected = self.driver.is_alive
        self.running = True

    def close(self):
        self.running = False
        self.connected = False
        if not self.driver:
            return
        self.driver.close()

    @wrapt.synchronized(lock)
    def reset(self):
        """Call this to force a rebuild/reconnect."""
        LOG.warning('Resetting client connection.')
        if self.driver:
            self.driver.close()
            if self.auto_connect:
                self.driver.setup_connection()
            if self.filter:
                self.driver.set_filter(self.filter)
        else:
            LOG.warning('Client not initialized, nothing to reset.')

    def send(self, packet: core.Packet) -> bool:
        if self.running:
            return self.driver.send(packet)
        else:
            LOG.error('Client not running, not sending packet.')
            return False

    # For the keepalive collector
    def keepalive_check(self):
        # Don't check the first time through.
        if not self.driver.is_alive and self._checks:
            LOG.warning("Resetting client.  It's not alive.")
            self.reset()
        self._checks = True

    # For the keepalive collector
    def keepalive_log(self):
        if ka := self.driver.keepalive:
            keepalive = timeago.format(ka)
        else:
            keepalive = 'N/A'
        LOGU.opt(colors=True).info(f'<green>Client keepalive {keepalive}</green>')

    def consumer(self, callback: Callable, raw: bool = False):
        if self.running:
            return self.driver.consumer(callback=callback, raw=raw)

    def decode_packet(self, *args, **kwargs) -> core.Packet:
        try:
            packet = self.driver.decode_packet(*args, **kwargs)
        except Exception as e:
            LOG.error(f'Error decoding packet: {e}')
            return None
        return packet
